// Author: Marek Handl
// Date: October 18, 2008
// CTU, FEE, X36PAA
// Knapsack problem - brute force and heuristic solution
// heuristic based on price/weight ratio

#include "stdafx.h"
#include <vector>
#include <sstream>
#include <iostream>
#include <string>
#include <fstream>
#include "assert.h"
#include "process.h"
#include "windows.h"
#include <algorithm>

using namespace std;


// ************************
// class representing one item
// ************************
class Item {
	int id;
	int price;
	int weight;
public:
	// constructor
	Item(int i, int p, int w) : id(i), price(p), weight(w) {}

	// operator less than (used for sorting)
	bool operator<(const Item& i) const {
		return ((float)price/(float)weight < (float)(i.price)/(float)(i.weight));
	}

	// getters
	int getPrice() {
		return price;
	}
	int getWeight() {
		return weight;
	}
};

// ************************
// class representing a knapsack
// ************************
class Knapsack {
	int maxWeight;
	int count;
	bool* content;
	int weight;
	int price;

public:
	Knapsack(int w, int c) : maxWeight(w), count(c), weight(0), price(0) {
		content = (count > 0 ? (bool*) malloc(count*sizeof(bool)) : NULL);
		for (int i=0; i<count; ++i) {
			content[i] = false;
		}
	}
	
	~Knapsack() {
		if (content != NULL) {
			free(content);
		}
	}
	
	Knapsack(const Knapsack& k) : maxWeight(k.maxWeight), count(k.count) {
		weight = k.weight;
		price = k.price;
		assert(count > 0);
		content = (count > 0 ? (bool*) malloc(count*sizeof(bool)) : NULL);
		for (int i=0; i<count; ++i) {
			content[i] = k.content[i];
		}
	}

	Knapsack& operator=(const Knapsack& k) {
		assert(count == k.count);
		maxWeight = k.maxWeight;
		weight = k.weight;
		price = k.price;
		for (int i=0; i<count; ++i) {
			assert(content != NULL);
			content[i] = k.content[i];
		}
		return *this;
	}

	// reset Knapsack (erase existing data, set new values)
	void reset(int w, int c) {
		maxWeight = w;
		weight = 0;
		price = 0;
		if (count != c) {	// in case count of items changes
			free(content);
			count = c;
			content = (count > 0 ? (bool*) malloc(count*sizeof(bool)) : NULL);
		}
		for (int i=0; i<count; ++i) {
			assert(content != NULL);
			content[i] = 0;
		}
	}
	
	// try to add an item into the sack
	// if it would exceed maxWeight, don't add and return false
	bool add(vector<Item> &items, int index) {
		assert(index < (int) items.size());
		if (weight + items.at(index).getWeight() > maxWeight || items.at(index).getPrice() == 0) {	// maximum weight overflow
			return false;
		}
		price += items.at(index).getPrice();
		weight += items.at(index).getWeight();
		content[index] = true;
		return true;
	}

	// getter
	int getPrice() const {
		return price;
	}

	// print knapsack into a file
	void printResult(ofstream &fout, int instanceId=0) {
		try {
			fout << instanceId << " " << count << " " << price << " ";
			for (int i=0; i<count; ++i) {
				fout << " " << (int) content[i];
			}
			fout << endl;
		}
		catch (...) {
			cout << "Output file error" << endl;
			exit(1);
		}
	}
	

	// output for debugging
	void printOut() const {
		cout << "Weight: " << weight << " (" << maxWeight << ")" << endl;
		cout << "Price: " << price << endl;
		for (int i=0; i<count; ++i) {
			cout << (int) content[i] << " ";
		}
	}
};


// ************************
// read input data from a file
// ************************
bool readLineFromFile(ifstream &fin, int &id, int &count, int &maxWeight, vector<Item> &items) {
	string line;
	bool readMore = true;
	id = count = maxWeight = 0;

	try {		
		if (fin.rdstate() & fstream::failbit) {		// if file not found or unable to read from it
			throw "Input file not found";
		}
		
		if (!(fin >> id >> count >> maxWeight)) {
			readMore = false;
		}
		else {
			items.reserve(count);
			for (int i=0; i<count; ++i) {
				int weight, price;
				fin >> weight >> price;
				Item item(i, price, weight);
				items.push_back(item);
			}
		}
		
	}
	catch (...) {
		fin.close();
		cout << "Input file error" << endl;
		exit(1);
	}
	return readMore;
}

// ************************
// logic of the program
// brute force - go through all possible combinations and find optimum
// omits obviously false combinations (when max weight already exceeded)
// ************************
void solve(Knapsack sack, vector<Item> &items, int index, Knapsack &best) {
	if (index >= (int) items.size()) {	// recursion depth constraint
		return;
	}
	
	solve(sack, items, index+1, best);	// recursion call - do not add current item into the sack
	
	if (sack.add(items, index)) {	// try to add current item into the sack
		solve(sack, items, index+1, best);	// recursion call
	}
	
	if (sack.getPrice() > best.getPrice()) {		// find the best solution
		best = sack;
	}
	// if you imagine all recursion calls as a binary tree...
	//   root represents start, its left child represents situation when the first item was not added
	//   its right child represents situation when the first item was added into the sack
	//   etc.
  // the price evaluation (the last if case) could be done in all leafes (with no children)
  //   then it would be placed in the recursion depth constraint if statement
  // current implementation does the check in all internal nodes
  //   it might seem it omits the last item, but that's not true
  //   it just doesn't differentiate between situation, when the last item gets and doesn't get added
  //   if it is possible (satisfies mass criterion) to add the last one, it computes the price, after it was added
  //   worth noting that the if statement can be true only in the nodes that are the ones before leafes
  //   "more" internal nodes can not obviously have a better price and they get computed later
  // note:
  //   count of leafes = 2^n
  //   count of internal nodes = 2^n - 1
  //   count of nodes, where the last if can be true = 2^(n-1)
}

// ************************
// heuristic solution
// sorts the items based on price/weight ratio in descending order
// ************************
void solveHeuristic(Knapsack &sack, vector<Item> &items) {
  sort(items.rbegin(), items.rend());
	for (int i=0; i<(int)items.size(); ++i) {
		sack.add(items, i);
	}
}

// ************************
// function used to measure time
// ************************
float GetMyCPUTime()
{
    HANDLE hProcess;    // specifies the process of interest
    FILETIME lpCreationTime;    // when the process was created
    FILETIME lpExitTime;    // when the process exited
    FILETIME lpKernelTime;    // time the process has spent in kernel mode
    FILETIME lpUserTime;
    LARGE_INTEGER t1, t2;


    hProcess = GetCurrentProcess();
    GetProcessTimes(hProcess, &lpCreationTime, &lpExitTime, &lpKernelTime, &lpUserTime);
    memcpy(&t1, &lpUserTime, sizeof(FILETIME));
    memcpy(&t2, &lpKernelTime, sizeof(FILETIME));
    return ((float)(t1.QuadPart+t2.QuadPart)/10000000);
}

// ************************
// ************************
int main(int argc, char* argv[])
{
	int instanceId;
	int count;
	int maxWeight;
	vector<Item> items;
	ifstream fin;
	ofstream fout, foutHeuristic;
	string outputFile("output.txt");
	string outputFileHeuristic("output1.txt");
	string inputFile("input.txt");
	float time = 0, timeHeuristic = 0, start, end;
	int instancesCount = 0;
	int maxInstancesCount = 0;	// zero means unlimited
	float error = 0, maxError = 0;

	if (argc < 2) {		// make sure program called with parameters
		cout << "Usage: knapsack.exe input_file [output_file [output_file_heuristic [maxInstancesCount]]]" << endl;
		return 0;
	}
	inputFile = string(argv[1]);
	if (argc > 2) {		// second parameter is optional and defines output file
		outputFile = string(argv[2]);
	}
	if (argc > 3) {		// second parameter is optional and defines output file
		outputFileHeuristic = string(argv[3]);
	}
	if (argc > 4) {
		maxInstancesCount = atoi(argv[4]);
	}

	fin.open(inputFile.c_str(), fstream::in);	// read only access
	fout.open(outputFile.c_str(), fstream::out);	// write only access
	foutHeuristic.open(outputFileHeuristic.c_str(), fstream::out);	// write only access
	
	Knapsack sack(1, 1);
	Knapsack best(1, 1);


	while (readLineFromFile(fin, instanceId, count, maxWeight, items) ) {
		// brute force
		sack.reset(maxWeight, count);
		best.reset(maxWeight, count);

		start = GetMyCPUTime();
		solve(sack, items, 0, best);	// compute solution
		end = GetMyCPUTime();

		time += end-start;
		best.printResult(fout, instanceId);		// save result to a file

		// heuristics
		sack.reset(maxWeight, count);
		start = GetMyCPUTime();
		solveHeuristic(sack, items);
		end = GetMyCPUTime();
		
		timeHeuristic += end-start;
		sack.printResult(foutHeuristic, instanceId);

		// comparison
		assert(sack.getPrice() <= best.getPrice());
		float currentError = (best.getPrice() - sack.getPrice()) / (float)(best.getPrice());
		error += currentError;
		maxError = (currentError > maxError ? currentError : maxError);

		items.clear();
		++instancesCount;
		if (instancesCount == maxInstancesCount) {	// stop execution if desired number of instances already computed
			break;
		}
	}
	
	fin.close();
	fout.close();

	cout << "Instances: " << instancesCount << endl;
	cout << "Time: " << time << " sec" << endl;
	cout << "Time per instance: " << time / (float)instancesCount << " sec" << endl;
	cout << "Time (heuristic): " << timeHeuristic << " sec" << endl;
	cout << "Time per instance (heuristic): " << timeHeuristic / (float)instancesCount << " sec" << endl;
	cout << "Average error: " << error / (float)instancesCount * 100 << "%" << endl;
	cout << "Maximum error: " << maxError*100 << "%" << endl;
	return 0;
}

