// X36PAA_Batoh.cpp : Defines the entry point for the console application.
// Author: Marek Handl
// Date: December 30, 2008
// CTU, FEE, X36PAA
// Knapsack problem - Dynamic programming approach

/* Dynamic programming
 * create table of dimensions: items_count x sum_of_items_costs
 * field (i,c) represents minimal weight, when only first i items used and sum of costs is c
*/

#include <vector>
#include <sstream>
#include <iostream>
#include <string>
#include <fstream>
#include "assert.h"
#include "process.h"
#include "windows.h"

using namespace std;

#define MAXINT 1000000

// ************************
// class representing one item
// ************************
class Item {
	int id;
	int price;
	int weight;
public:
	Item(int i, int p, int w) : id(i), price(p), weight(w) {}

	// getters
	int getPrice() {
		return price;
	}
	int getWeight() {
		return weight;
	}
};

// ************************
// class representing a knapsack
// ************************
class Knapsack {
	int maxWeight;
	int count;
	bool* content;
	int weight;
	int price;

public:
	Knapsack(int w, int c) : maxWeight(w), count(c), weight(0), price(0) {
		content = (count > 0 ? (bool*) malloc(count*sizeof(bool)) : NULL);
		for (int i=0; i<count; ++i) {
			content[i] = false;
		}
	}
	~Knapsack() {
		if (content != NULL) {
			free(content);
		}
	}
	Knapsack(const Knapsack& k) : maxWeight(k.maxWeight), count(k.count) {
		weight = k.weight;
		price = k.price;
		assert(count > 0);
		content = (count > 0 ? (bool*) malloc(count*sizeof(bool)) : NULL);
		for (int i=0; i<count; ++i) {
			content[i] = k.content[i];
		}
	}

	Knapsack& operator=(const Knapsack& k) {
		assert(count == k.count);
		maxWeight = k.maxWeight;
		weight = k.weight;
		price = k.price;
		for (int i=0; i<count; ++i) {
			assert(content != NULL);
			content[i] = k.content[i];
		}
		return *this;
	}
	// no need for operator= or copy constructor, everything is static

	// reset Knapsack (erase existing data, set new values)
	void reset(int w, int c) {
		maxWeight = w;
		weight = 0;
		price = 0;
		if (count != c) {	// in case count of items changes
			free(content);
			count = c;
			content = (count > 0 ? (bool*) malloc(count*sizeof(bool)) : NULL);
		}
		for (int i=0; i<count; ++i) {
			assert(content != NULL);
			content[i] = 0;
		}
	}
	
	// try to add an item into the sack
	// if it would exceed maxWeight, don't add and return false
	bool add(vector<Item> &items, int index) {
		assert(index < (int) items.size());
		if (weight + items.at(index).getWeight() > maxWeight) {	// maximum weight overflow
			return false;
		}
		price += items.at(index).getPrice();
		weight += items.at(index).getWeight();
		content[index] = true;
		return true;
	}

	// remove item from the sack
	void remove(vector<Item> &items, int index) {
		assert(index < (int) items.size());
		assert(content[index]);	// make sure the item is in the sack
		price -= items.at(index).getPrice();
		weight -= items.at(index).getWeight();
		content[index] = false;
	}

	// getter
	int getPrice() const {
		return price;
	}

	// print knapsack into a file
	void printResult(ofstream &fout, int instanceId=0) {
		try {
			fout << instanceId << " " << count << " " << price << " ";
			for (int i=0; i<count; ++i) {
				fout << " " << (int) content[i];
			}
			fout << endl;
		}
		catch (...) {
			cout << "Output file error" << endl;
			exit(1);
		}
	}
	

	// output for debugging
	void printOut() const {
		cout << "Weight: " << weight << " (" << maxWeight << ")" << endl;
		cout << "Price: " << price << endl;
		for (int i=0; i<count; ++i) {
			cout << (int) content[i] << " ";
		}
	}
};


// ************************
// read input data from a file
// ************************
bool readLineFromFile(ifstream &fin, int &id, int &count, int &maxWeight, vector<Item> &items) {
	string line;
	bool readMore = true;
	id = count = maxWeight = 0;

	try {		
		if (fin.rdstate() & fstream::failbit) {		// if file not found or unable to read from it
			throw "Input file not found";
		}
		
		if (!(fin >> id >> count >> maxWeight)) {
			readMore = false;
		}
		else {
			items.reserve(count);
			for (int i=0; i<count; ++i) {
				int weight, price;
				fin >> weight >> price;
				Item item(i, price, weight);
				items.push_back(item);
			}
		}
		
	}
	catch (...) {
		fin.close();
		cout << "Input file error" << endl;
		exit(1);
	}
	return readMore;
}

// ************************
// logic of the program
// ************************
void fillTable(Knapsack sack, vector<Item> &items, int* table, int count, int costSum) {
	
	// i - row (sum of costs)
	// j - column (items)
	for (int i=1; i<=costSum; ++i) {
		for (int j=1; j<=count; ++j) {
			// W(j+1,c) = min{ W(j,c) , W(j,c-c[j+1])+w[j+1]) } 
			// i.e. W(j,c) = min{ W(j-1,c) , W(j-1,c-c[j])+w[j]) } 

			int valueOne = table[i*(count+1)+(j-1)];	// W(j-1,c)

			int valueTwo = MAXINT;
			
			int index = i - items[j-1].getPrice();	// c-c[j]	... items are indexed from 0, therefore [j-1]
			if (index >= 0) {
				valueTwo = table[index*(count+1)+(j-1)] + items[j-1].getWeight();	// W(j-1,c-c[j])+w[j]	... items are indexed from 0, therefore [j-1]
			}
			else {
				// don't do anything, leave infinity there
			}
			
			table[i*(count+1)+j] = valueOne < valueTwo ? valueOne : valueTwo;
		}
	}
}


// ************************
// function used to measure time
// ************************
float GetMyCPUTime()
{
    HANDLE hProcess;    // specifies the process of interest
    FILETIME lpCreationTime;    // when the process was created
    FILETIME lpExitTime;    // when the process exited
    FILETIME lpKernelTime;    // time the process has spent in kernel mode
    FILETIME lpUserTime;
    LARGE_INTEGER t1, t2;

    hProcess = GetCurrentProcess();
    GetProcessTimes(hProcess, &lpCreationTime, &lpExitTime, &lpKernelTime, &lpUserTime);
    memcpy(&t1, &lpUserTime, sizeof(FILETIME));
    memcpy(&t2, &lpKernelTime, sizeof(FILETIME));
    return ((float)(t1.QuadPart+t2.QuadPart)/10000000);
}

// ************************
// get the result - best value and what items to add to the knapsack
// the result is returned in the sack reference variable
// ************************
void findResult(const int *table, int count, int costSum, int maxWeight, Knapsack &sack, vector<Item> &items) {
	int value = 0;
	int indexOfResult = 0;
	// i - row (sum of costs)
	// j - column (items)
	
	// first find the final result, i.e. what cost is the highest achievable
	for (int i=costSum; i>=0; --i) {	// use descending order
		value = table[i*(count+1)+count];	// check the last column in rows
		if (value <= maxWeight) {
			indexOfResult = i;
			break;
		}
	}

	// then find what items build up the result
	int i = indexOfResult;
	int j = count-1;
	while (j>=0) {
		if (table[i*(count+1)+j] != table[i*(count+1)+j+1]) {	// if value to the left is different, this item is part of the solution
			sack.add(items, j);			// note - items are indexed from 0, table items from 1
			i -= items[j].getPrice();	
			--j;			
		}
		else {
			while (j>=0 && table[i*(count+1)+j] == table[i*(count+1)+j+1]) {	// if the values are equal, this item is not part of the solution
				--j;
			}
		}
	}
}

// ************************
// allocate memory for results table
// ************************
int* getTable(int count, int costSum) {
	int *table;
	table = (int*) malloc(((count+1)*(costSum+1)) * sizeof(int));
	// i - row (sum of costs)
	// j - column (items)

	// first row contains zeros
	for (int j=0; j<=count; ++j) {
		table[0+j] = 0;
	}
	// other rows contain infinity
	for (int i=1; i<=costSum; ++i) {
		for (int j=0; j<=count; ++j) {
			table[i*(count+1)+j] = MAXINT;
		}
	}
	return table;
}

// ************************
// ************************
int main(int argc, char* argv[])
{
	int instanceId;
	int count;
	int maxWeight;
	vector<Item> items;
	ifstream fin;
	ofstream fout;
	string outputFile("output.txt");
	string inputFile("input.txt");
	float time = 0, start, end;
	int instancesCount = 0;
	int *table = NULL;
	
	if (argc < 2) {		// make sure program is called with parameters
		cout << "Usage: X36PAA_Batoh.exe input_file [output_file]" << endl;
		return 0;
	}
	inputFile = string(argv[1]);
	if (argc > 2) {		// second parameter is optional and defines output file
		outputFile = string(argv[2]);
	}

	fin.open(inputFile.c_str(), fstream::in);	// read only access
	fout.open(outputFile.c_str(), fstream::out);	// write only access
	
	Knapsack sack(1, 1);
	Knapsack best(1, 1);

	while (readLineFromFile(fin, instanceId, count, maxWeight, items) ) {
		sack.reset(maxWeight, count);
		best.reset(maxWeight, count);
		
		// get theoretical max (sum costs of all items)
		int costSum = 0;
		for (int i=0; i<count; ++i) {
			costSum += items[i].getPrice();
		}

		// allocate memory for table that will contain results
		if (table != NULL) {
			free(table);
		}
		table = getTable(count, costSum);	// allocate memory for table

		start = GetMyCPUTime();
		fillTable(sack, items, table, count, costSum);		// populate fields of the table
		findResult(table, count, costSum, maxWeight, best, items);	// find solution
		end = GetMyCPUTime();

		time += end-start;
		++instancesCount;
		best.printResult(fout, instanceId);		// save result to a file
		items.clear();
	}
	
	fin.close();
	fout.close();	

	cout << "Instances: " << instancesCount << endl;
	cout << "Time: " << time << " sec" << endl;
	cout << "Time per instance: " << (float)time / (float)instancesCount << " sec" << endl;
	return 0;
}

